@@ -1,5 +1,8 @@
+use crate::car::Car;
 use pyo3::exceptions;
 
+pub const NEW_CAR: Car = Car::new();
+
 pub const MAX_SPEED: f32 = 2300.;
 pub const MAX_SPEED_NO_BOOST: f32 = 1410.;
 pub const MIN_SPEED: f32 = -MAX_SPEED_NO_BOOST;
